require_relative '../lib/named_parameters'


class Item

  attr_reader :id

  def setup(options = { })
    @id = options[:id]
  end

  def setup_only_valid_attributes(options = { })
    @id = options[:id]
  end

end


class Book < Item

  include NamedParameters

  attr_reader :author, :title


  named_parameters(author: 'unknown')

  def edit_attributes(author, title)
    @author = author
    @title  = title
  end


  named_parameters_strict(author: 'unknown')

  def edit_only_valid_attributes(author, title)
    @author = author if author.is_a? String
    @title = title if title.is_a? String
  end


  named_parameters_super(author: 'unknown')

  def setup(author, title)
    edit_attributes(author: author,
                    title:  title)
  end

  named_parameters_strict_super(author: 'unknown')

  def setup_only_valid_attributes(author, title, id)
    edit_attributes(author: author,
                    title:  title)
  end

  named_parameters(attributes: [])

  def change_attributes(attributes)
    attributes.each do |attribute|
      instance_variable_set("@#{attribute}", yield(send(attribute)))
    end
  end
end

class Magazine < Item
  include NamedParameters

  attr_reader :author, :title

  def edit_attributes(author, title)
    @author = author
    @title  = title
  end

  edit_attributes_has_named_parameters(author: 'unknown')


  def edit_only_valid_attributes(author, title)
    @author = author if author.is_a? String
    @title = title if title.is_a? String
  end

  edit_only_valid_attributes_has_named_parameters_strict(author: 'unknown')


  def setup(author, title)
    edit_attributes(author: author,
                    title:  title)
  end

  setup_has_named_parameters_super(author: 'unknown')

  def setup_only_valid_attributes(author, title, id)
    edit_attributes(author: author,
                    title:  title)
  end

  setup_only_valid_attributes_has_named_parameters_strict_super(author: 'unknown')

  def change_attributes(attributes)
    attributes.each do |attribute|
      instance_variable_set("@#{attribute}", yield(send(attribute)))
    end
  end

  change_attributes_has_named_parameters(attributes: [])

  def check_for_attributes(author, title)
    @author == author && @title == title
  end

  check_for_attributes_has_named_parameters
end

describe NamedParameters do

  attr_reader :book, :magazine
  before :each do
    @book     = Book.new
    @magazine = Magazine.new
  end

  describe 'default' do
    it 'should work with all parameters given' do
      book.edit_attributes(author: 'Orson Scott Card',
                           title:  'Enders Game')
      book.title.should == 'Enders Game'
      book.author.should == 'Orson Scott Card'
    end

    it 'should work with only mandatory parameters given' do
      book.edit_attributes(title: 'Enders Game')
      book.title.should == 'Enders Game'
      book.author.should == 'unknown'
    end

    it 'should throw an error when a mandatory parameter misses' do
      -> { book.edit_attributes(author: 'Orson Scott Card') }.should raise_error(RuntimeError)
    end

    it 'should ignore options that are not part of the parameters' do
      book.edit_attributes(author: 'Orson Scott Card',
                           title:  'Enders Game',
                           year:   1985)
      book.title.should == 'Enders Game'
      book.author.should == 'Orson Scott Card'
    end
  end


  describe 'strict named parameters' do

    it 'should work with all parameters given' do
      book.edit_only_valid_attributes(author: 'Orson Scott Card',
                                      title:  'Enders Game')
      book.title.should == 'Enders Game'
      book.author.should == 'Orson Scott Card'
    end

    it 'should work with only mandatory parameters given' do
      book.edit_only_valid_attributes(title: 'Enders Game')
      book.title.should == 'Enders Game'
      book.author.should == 'unknown'
    end

    it 'should throw an error when a mandatory parameter misses' do
      -> { book.edit_only_valid_attributes(author: 'Orson Scott Card') }.should raise_error(RuntimeError)
    end

    it 'should throw an error when there are options that are not part of the parameters' do
      -> { book.edit_only_valid_attributes(author: 'Orson Scott Card',
                                           title:  'Enders Game',
                                           year:   1985) }.should raise_error(RuntimeError)
    end

  end


  describe 'with super' do

    it 'should call super with all options as the only parameter' do
      book.setup(author: 'Orson Scott Card',
                 title:  'Enders Game',
                 year:   1985,
                 id:     1)
      book.title.should == 'Enders Game'
      book.author.should == 'Orson Scott Card'
      book.id.should == 1
    end

  end

  describe 'with super and strict' do

    it 'should call super' do
      book.setup_only_valid_attributes(author: 'Orson Scott Card',
                                       title:  'Enders Game',
                                       id:     1)
      book.title.should == 'Enders Game'
      book.author.should == 'Orson Scott Card'
      book.id.should == 1
    end

    it 'should throw error when there are unwanted options' do
      -> { book.setup_only_valid_attributes(author: 'Orson Scott Card',
                                            title:  'Enders Game',
                                            year:   1985,
                                            id:     1) }.should raise_error(RuntimeError)
    end

  end

  describe 'with block' do
    it 'should forward blocks' do
      book.edit_attributes(author: 'Orson Scott Card',
                           title:  'Enders Game')
      book.change_attributes(attributes: [:author, :title]) do |attribute|
        attribute.to_s.upcase
      end
      book.author.should == 'ORSON SCOTT CARD'
      book.title.should == 'ENDERS GAME'
    end
  end

  describe 'ghost_methods' do

    describe 'default' do
      it 'should work with all parameters given' do
        magazine.edit_attributes(author: 'Orson Scott Card',
                                 title:  'Enders Game')
        magazine.title.should == 'Enders Game'
        magazine.author.should == 'Orson Scott Card'
      end

      it 'should work with only mandatory parameters given' do
        magazine.edit_attributes(title: 'Enders Game')
        magazine.title.should == 'Enders Game'
        magazine.author.should == 'unknown'
      end

      it 'should throw an error when a mandatory parameter misses' do
        -> { magazine.edit_attributes(author: 'Orson Scott Card') }.should raise_error(RuntimeError)
      end

      it 'should ignore options that are not part of the parameters' do
        magazine.edit_attributes(author: 'Orson Scott Card',
                                 title:  'Enders Game',
                                 year:   1985)
        magazine.title.should == 'Enders Game'
        magazine.author.should == 'Orson Scott Card'
      end
    end

    describe 'strict named parameters' do

      it 'should work with all parameters given' do
        magazine.edit_only_valid_attributes(author: 'Orson Scott Card',
                                            title:  'Enders Game')
        magazine.title.should == 'Enders Game'
        magazine.author.should == 'Orson Scott Card'
      end

      it 'should work with only mandatory parameters given' do
        magazine.edit_only_valid_attributes(title: 'Enders Game')
        magazine.title.should == 'Enders Game'
        magazine.author.should == 'unknown'
      end

      it 'should throw an error when a mandatory parameter misses' do
        -> { magazine.edit_only_valid_attributes(author: 'Orson Scott Card') }.should raise_error(RuntimeError)
      end

      it 'should throw an error when there are options that are not part of the parameters' do
        -> { magazine.edit_only_valid_attributes(author: 'Orson Scott Card',
                                                 title:  'Enders Game',
                                                 year:   1985) }.should raise_error(RuntimeError)
      end

    end


    describe 'with super' do

      it 'should call super with all options as the only parameter' do
        magazine.setup(author: 'Orson Scott Card',
                       title:  'Enders Game',
                       year:   1985,
                       id:     1)
        magazine.title.should == 'Enders Game'
        magazine.author.should == 'Orson Scott Card'
        magazine.id.should == 1
      end

    end

    describe 'with super and strict' do

      it 'should call super' do
        magazine.setup_only_valid_attributes(author: 'Orson Scott Card',
                                             title:  'Enders Game',
                                             id:     1)
        magazine.title.should == 'Enders Game'
        magazine.author.should == 'Orson Scott Card'
        magazine.id.should == 1
      end

      it 'should throw error when there are unwanted options' do
        -> { magazine.setup_only_valid_attributes(author: 'Orson Scott Card',
                                                  title:  'Enders Game',
                                                  year:   1985,
                                                  id:     1) }.should raise_error(RuntimeError)
      end

    end

    describe 'with block' do
      it 'should forward blocks' do
        magazine.edit_attributes(author: 'Orson Scott Card',
                                 title:  'Enders Game')
        magazine.change_attributes(attributes: [:author, :title]) do |attribute|
          attribute.to_s.upcase
        end
        magazine.author.should == 'ORSON SCOTT CARD'
        magazine.title.should == 'ENDERS GAME'
      end
    end

    describe 'missing methods' do
      it 'should not catch any other missing methods' do
        -> { magazine.some_weird_method }.should raise_error(NoMethodError)
      end
    end

    it 'should handle no optionals' do
      -> { magazine.check_for_attributes(author: 'Orson Scott Card') }.should raise_error(RuntimeError)
      magazine.setup(author: 'Orson Scott Card',
                     title:  'Enders Game')
      magazine.check_for_attributes(author: 'Orson Scott Card',
                                    title:  'Enders Game').should be_true
    end
  end

end
